
# Python3 processing script for processing responses to CALL shared task
#
# The script is based on the minimal python script supplied with the text task,
# but instead of using the grammar to decide on accepting/rejecting, reads the 
# decision from the answer sheet.

# The sheet with the gold standard answers is specified by 'gs_spreadsheet'

# The sheets to process are specified at the end, in the calls to 'process_answer_sheet'

# ---------------------------------------------------

# Example of running the script (all files in same directory as script):

# $ python3 score_test_data_answers.py
# Read answer spreadsheet: sample_judgements.csv
# Reading and processing judgements spreadsheet: textProcessing_trainingKaldi_20170213.csv
# 
# INCORRECT UTTERANCES (1342)
# CorrectReject    1268
# GrossFalseAccept 20*3 = 60
# PlainFalseAccept 54
# RejectionRate    0.92
# 
# CORRECT UTTERANCES (3880)
# CorrectAccept    2154
# FalseReject      1726
# RejectionRate    0.44
# 
# D                2.06
# 
# Written scored spreadsheet: sample_judgements_with_scores.csv

# ---------------------------------------------------

# Weighting factor for gross false accepts

k = 3

# Files used:

# gs_spreadsheet is the input spreadsheet with the prompts, recognition results, transcriptions and judgements

# For training data
#gs_spreadsheet = 'textProcessing_trainingKaldi_20170213.csv'

# For test data
gs_spreadsheet = 'ST3TestDataGoldStandard.csv'

# ---------------------------------------------------

import csv
import sys
import math

# ---------------------------------------------------

def process_answer_sheet(spreadsheet_in, modality, spreadsheet_out):
    print("\n---------------------------------\n")
    print("Read answer spreadsheet: " + spreadsheet_in)
    answer_dic = read_answer_spreadsheet(spreadsheet_in, modality)
    print("Reading and processing judgements spreadsheet: " + gs_spreadsheet)
    read_and_process_spreadsheet(answer_dic, gs_spreadsheet, spreadsheet_out)
    print("\nWritten scored spreadsheet: " + spreadsheet_out)

# ---------------------------------------------------

# Start off by reading the answer spreadsheet into a dictionary 
# which associates ids with judgement
#
# Return the dictionary

def read_answer_spreadsheet(spreadsheet_in, modality):
    check_modality(modality)
    dictionary = {}
    with open(spreadsheet_in, 'r', encoding="utf-8") as csv_infile:
        reader = csv.reader(csv_infile, delimiter='\t', quotechar='"')
        for row in reader:
            # Skip header row
            if ( not is_header_row(row) and modality == 'speech' and len(row) >= 4 ):
                id = row[0]
                judgement = row[3]
                dictionary[id] = judgement.lower()
            elif ( not is_header_row(row) and modality == 'text' and len(row) >= 5 ):
                id = row[0]
                judgement = row[4]
                dictionary[id] = judgement.lower()
    return dictionary

def check_modality(modality):
    if ( modality != 'text' and modality != 'speech' ):
        print("\nUnknown modality: " + modality)
        sys.exit()        

# ---------------------------------------------------

# Open the input and output files and process each row, skipping the header
def read_and_process_spreadsheet(answer_dic, gs_spreadsheet, spreadsheet_out):
    out_fieldnames = ['Id', 'Prompt', 'Transcription', 'RecResult', 'Language', 'Meaning', 'Accept', 'Result']
    scores = init_scores()
    with open(gs_spreadsheet, 'r', encoding="utf-8") as csv_infile:
        reader = csv.reader(csv_infile, delimiter='\t', quotechar='"')
        with open(spreadsheet_out, 'w', encoding="utf-8") as csv_outfile:
            writer = csv.DictWriter(csv_outfile, fieldnames=out_fieldnames, delimiter='\t', quotechar='"', lineterminator="\n")
            writer.writeheader()
            for row in reader:
                # Skip header row
                if ( not is_header_row(row) ):
                    process_spreadsheet_row(row, answer_dic, writer, scores)
    print_scores(scores)

def init_scores():
    return {'CorrectAccept': 0, 'GrossFalseAccept': 0, 'PlainFalseAccept': 0, 'CorrectReject': 0, 'FalseReject': 0}

def is_header_row(row):
    return ( row[0] == 'Id' )

def process_spreadsheet_row(row, answer_dic, writer, scores):
    id = row[0]
    prompt = row[1]
    rec_result = row[2]
    transcription = row[3]
    language_correct_gold_standard = row[4]
    meaning_correct_gold_standard = row[5]
    decision = classification_according_to_answer_spreadsheet(id, answer_dic)
    result = score_decision(decision, language_correct_gold_standard, meaning_correct_gold_standard, scores)

    writer.writerow({'Id': id, 
                     'Prompt': prompt, 
                     'RecResult': rec_result,
                     'Transcription': transcription,
                     'Language': language_correct_gold_standard, 
                     'Meaning': meaning_correct_gold_standard, 
                     'Accept': decision, 
                     'Result': result})

# Look up <id, judgement> pair in dictionary.
# Print an error if the id isn't mentioned in the dictionary.
def classification_according_to_answer_spreadsheet(id, answer_dic):
    if ( id in answer_dic.keys() ):
        return answer_dic[id]
    else:
        print("*** Error: id not in gold standard spreadsheet: '" + id + "'") 
        return 'UNDEFINED'

# Compare decision with gold standard judgements for language and meaning
def score_decision(decision, language_correct_gs, meaning_correct_gs, scores):
    if ( decision == 'accept' and language_correct_gs == 'correct' ):
        result = 'CorrectAccept'
    elif ( decision == 'accept' and meaning_correct_gs == 'incorrect' ):
        result = 'GrossFalseAccept'
    elif ( decision == 'accept' ):
        result = 'PlainFalseAccept'
    elif ( decision == 'reject' and language_correct_gs == 'incorrect' ):
        result = 'CorrectReject'
    else:
        result = 'FalseReject'
    scores[result] = scores[result] + 1
    return result

def print_scores(scores):
    CA = scores['CorrectAccept']
    GFA = scores['GrossFalseAccept']
    PFA = scores['PlainFalseAccept']
    CR = scores['CorrectReject']
    FR = scores['FalseReject']

    FA = PFA + k * GFA
    Correct = CA + FR
    Incorrect = CR + GFA + PFA
    
    if ( CR + FA ) > 0 :
        IncorrectRejectionRate = CR / ( CR + FA )
    else:
        IncorrectRejectionRate = 'undefined'

    if ( CR + FA ) > 0 :
        IncorrectAcceptanceRate = FA / ( CR + FA )
    else:
        IncorrectAcceptanceRate = 'undefined'

    if ( FR + CA ) > 0 :
        CorrectRejectionRate = FR / ( FR + CA )
    else:
        CorrectRejectionRate = 'undefined'

    if ( FR + CA ) > 0 :
        CorrectAcceptanceRate = CA / ( FR + CA )
    else:
        CorrectAcceptanceRate = 'undefined'

    if ( CorrectRejectionRate != 'undefined' and IncorrectRejectionRate != 'undefined' ) :
        D = IncorrectRejectionRate / CorrectRejectionRate 
    else:
        D = 'undefined'

    if ( CorrectAcceptanceRate != 'undefined' and IncorrectAcceptanceRate != 'undefined' ) :
        DA = CorrectAcceptanceRate / IncorrectAcceptanceRate
    else:
        DA = 'undefined'

    if ( D != 'undefined' and DA != 'undefined' ):
        D_FULL = math.sqrt( D * DA )
    else:
        D_FULL = 'undefined' 

    print('\nINCORRECT UTTERANCES (' + str(Incorrect) + ')' )
    print('CorrectReject    ' + str(CR) )
    print('GrossFalseAccept ' + str(GFA) + '*' + str(k) + ' = ' + str(GFA * k) )
    print('PlainFalseAccept ' + str(PFA) )
    print('AcceptanceRate   ' + two_digits(IncorrectAcceptanceRate) )
    print('RejectionRate    ' + two_digits(IncorrectRejectionRate) )

    print('\nCORRECT UTTERANCES (' + str(Correct) + ')')
    print('CorrectAccept    ' + str(CA) )
    print('FalseReject      ' + str(FR) )
    print('AcceptanceRate   ' + two_digits(CorrectAcceptanceRate) )
    print('RejectionRate    ' + two_digits(CorrectRejectionRate) )

    print('\nD                ' + two_digits(D) )
    print('\nDA               ' + two_digits(DA) )
    print('\nD_FULL           ' + two_digits(D_FULL) )

def two_digits(x):
    if x == 'undefined':
        return 'undefined'
    else:
        return ( "%.2f" % x )

# ---------------------------------------------------

# For training data
#process_answer_sheet('sample_judgements.csv', 'speech', 'sample_judgements_with_scores.csv')

# For test data
process_answer_sheet('AAA_SpeechTask.csv', 'speech', 'AAA_SpeechTask_with_scores.csv')
process_answer_sheet('BBB_SpeechTask.csv', 'speech', 'BBB_SpeechTask_with_scores.csv')
process_answer_sheet('CCC_SpeechTask.csv', 'speech', 'CCC_SpeechTask_with_scores.csv')

process_answer_sheet('DDD_TextTask.csv', 'text', 'DDD_TextTask_with_scores.csv')
process_answer_sheet('EEE_TextTask.csv', 'text', 'EEE_TextTask_with_scores.csv')
process_answer_sheet('FFF_TextTask.csv', 'text', 'FFF_TextTask_with_scores.csv')

process_answer_sheet('GGG_SpeechTask.csv', 'speech', 'GGG_SpeechTask_with_scores.csv')
process_answer_sheet('HHH_SpeechTask.csv', 'speech', 'HHH_SpeechTask_with_scores.csv')
process_answer_sheet('III_SpeechTask.csv', 'speech', 'III_SpeechTask_with_scores.csv')

process_answer_sheet('JJJ_TextTask.csv', 'text', 'JJJ_TextTask_with_scores.csv')

process_answer_sheet('KKK_TextTask.csv', 'text', 'KKK_TextTask_with_scores.csv')
process_answer_sheet('LLL_TextTask.csv', 'text', 'LLL_TextTask_with_scores.csv')
process_answer_sheet('MMM_TextTask.csv', 'text', 'MMM_TextTask_with_scores.csv')

process_answer_sheet('Baseline_TextTask.csv', 'text', 'Baseline_TextTask_with_scores.csv')
process_answer_sheet('BaselinePerfectRec_TextTask.csv', 'text', 'BaselinePerfectRec_TextTask_with_scores.csv')
